/* Intel and AMD x86 CPUID display program
 * Copyright 1999 Phil Karn, KA9Q
 * May be used under the terms of the GNU Public License (GPL)

 * Reference documents:
 * ftp://download.intel.com/design/pro/applnots/24161809.pdf  (AP-485)
 * http://developer.intel.com/design/pentiumiii/applnots/24512501.pdf (AP-909)
 * http://www.amd.com/K6/k6docs/pdf/20734j.pdf
 */

#include <stdio.h>

void cpuid(long *,long *,long *,long *);
void decode_intel_tlb(int x);

main()
{
  int i,maxi;
  unsigned long eax,ebx,ecx,edx;

  /* Dump all the CPUID results in raw hex */
  eax = 0;
  cpuid(&eax,&ebx,&ecx,&edx);
  maxi = eax;
  for(i=0;i<=maxi;i++){
    eax = i;
    cpuid(&eax,&ebx,&ecx,&edx);
    printf("eax in: %i, eax = %08x ebx = %08x ecx = %08x edx = %08x\n",i,eax,ebx,ecx,edx);
  }

  /* Vendor ID and max CPUID level supported */
  eax = 0;
  cpuid(&eax,&ebx,&ecx,&edx);
  printf("Vendor ID: \"");
  for(i=0;i<4;i++)
    putchar(ebx >> (8*i));
  for(i=0;i<4;i++)
    putchar(edx >> (8*i));
  for(i=0;i<4;i++)
    putchar(ecx >> (8*i));
  printf("\"; CPUID level %d\n\n",maxi);

  switch(ebx){
  case 0x756e6547: /* Intel */
    dointel();
    break;
  case 0x68747541: /* AMD */
    doamd();
    break; /* Need Cyrix in here too */
  default:
    printf("Unknown vendor\n");
    break;
  }
}

/* Intel-specific information */
dointel(){
  int i,maxi;
  unsigned long eax,ebx,ecx,edx;
  int stepping,model,family,type,reserved,ntlb;

  printf("Intel-specific functions\n");
  eax = 0;
  cpuid(&eax,&ebx,&ecx,&edx);
  maxi = eax;


  if(maxi >= 1){
    /* Family/model/type etc */
    eax = 1;
    cpuid(&eax,&ebx,&ecx,&edx);
    stepping = eax & 0xf;
    model = (eax >> 4) & 0xf;
    family = (eax >> 8) & 0xf;
    type = (eax >> 12) & 0x3;
    reserved = eax >> 14;

    printf("Family: %d Model: %d Type %d [",family,model,type);
    switch(family){
    case 3:
      printf("i386 ");
      switch(model){
      }
      break;
    case 4:
      printf("i486 ");
      switch(model){
      case 0:
      case 1:
	printf("DX");
	break;
      case 2:
	printf("SX");
	break;
      case 3:
	printf("487/DX2");
	break;
      case 4:
	printf("SL");
	break;
      case 5:
	printf("SX2");
	break;
      case 7:
	printf("write-back enhanced DX2");
	break;
      case 8:
	printf("DX4");
	break;
      }
      break;
    case 5:
      printf("Pentium ");
      switch(model){
      case 1:
	printf("60/66");
	break;
      case 2:
	printf("75-200");
	break;
      case 3:
	printf("for 486 system");
	break;
      case 4:
	printf("MMX");
	break;
      }
      break;
    case 6:
      printf("Pentium Pro; ");
      switch(model){
      case 1:
	printf("Pentium Pro;");
	break;
      case 3:
	printf("Pentium II Model 3;");
	break;
      case 5:
	printf("Pentium II Model 5/Xeon/Celeron;");
	break;
      case 6:
	printf("Celeron;");
	break;
      case 7:
	printf("Pentium III/Pentium III Xeon;");
	break;
      }
      break;
    }
    switch(type){
    case 0:
      printf(" Original OEM");
      break;
    case 1:
      printf(" Overdrive");
      break;
    case 2:
      printf(" Dual-capable");
      break;
    case 3:
      printf(" Reserved");
      break;
    }
    printf("]\n");
    printf("Stepping: %d\n",stepping);

    printf("Reserved: %d\n\n",reserved);

    printf("Feature flags:\n");
    if(edx & (1<<0))
      printf("FPU - Floating Point Unit\n");
    if(edx & (1<<1))
      printf("VME - Virtual 8086 Mode Enhancements\n");
    if(edx & (1<<2))
      printf("DE - Debugging Extensions\n");
    if(edx & (1<<3))
      printf("PSE - Page Size Extensions\n");
    if(edx & (1<<4))
      printf("TSC - Time Stamp Counter\n");
    if(edx & (1<<5))
      printf("MSR - Model Specific Registers\n");
    if(edx & (1<<6))
      printf("PAE - Physical Address Extension\n");
    if(edx & (1<<7))
      printf("MCE - Machine Check Exception\n");
    if(edx & (1<<8))
      printf("CX8 - COMPXCHG8B Instruction\n");
    if(edx & (1<<9))
      printf("APIC - On-chip Advanced Programmable Interrupt Controller present and enabled\n");
    if(edx & (1<<10))
      printf("10 - Reserved\n");
    if(edx & (1<<11))
      printf("SEP - Fast System Call\n");
    if(edx & (1<<12))
      printf("MTRR - Memory Type Range Registers\n");
    if(edx & (1<<13))
      printf("PGE - PTE Global Flag\n");
    if(edx & (1<<14))
      printf("MCA - Machine Check Architecture\n");
    if(edx & (1<<15))
      printf("CMOV - Conditional Move and Compare Instructions\n");
    if(edx & (1<<16))
      printf("FGPAT - Page Attribute Table\n");
    if(edx & (1<<17))
      printf("PSE-36 - 36-bit Page Size Extension\n");
    if(edx & (1<<18))
      printf("PN - Processor Serial Number present and enabled\n");
    if(edx & (1<<19))
      printf("19 - reserved\n");
    if(edx & (1<<20))
      printf("20 - reserved\n");
    if(edx & (1<<21))
      printf("21 - reserved\n");
    if(edx & (1<<22))
      printf("22 - reserved\n");
    if(edx & (1<<23))
      printf("MMX - MMX instruction set\n");
    if(edx & (1<<24))
      printf("FXSR - Fast FP/MMX Streaming SIMD Extensions save/restore\n");
    if(edx & (1<<25))
      printf("XMM - Streaming SIMD Extensions instruction set\n");
    if(edx & (1<<26))
      printf("26 - reserved\n");
    if(edx & (1<<27))
      printf("27 - reserved\n");
    if(edx & (1<<28))
      printf("28 - reserved\n");
    if(edx & (1<<29))
      printf("29 - reserved\n");
    if(edx & (1<<30))
      printf("30 - reserved\n");
    if(edx & (1<<31))
      printf("31 - reserved\n");
  }

  if(maxi >= 2){
    /* Decode TLB and cache info */
    ntlb = 255;
    for(i=0;i<ntlb;i++){
      eax = 2;
      cpuid(&eax,&ebx,&ecx,&edx);
      ntlb =  eax & 0xff;
      decode_intel_tlb(eax >> 8);
      decode_intel_tlb(eax >> 16);
      decode_intel_tlb(eax >> 24);
      
      if((ebx & 0x80000000) == 0){
	decode_intel_tlb(ebx);
	decode_intel_tlb(ebx >> 8);
	decode_intel_tlb(ebx >> 16);
	decode_intel_tlb(ebx >> 24);
      }
      if((ecx & 0x80000000) == 0){
	decode_intel_tlb(ecx);
	decode_intel_tlb(ecx >> 8);
	decode_intel_tlb(ecx >> 16);
	decode_intel_tlb(ecx >> 24);
      }
      if((edx & 0x80000000) == 0){
	decode_intel_tlb(edx);
	decode_intel_tlb(edx >> 8);
	decode_intel_tlb(edx >> 16);
	decode_intel_tlb(edx >> 24);
      }
    }
  }
  if(maxi >= 3){
    /* Pentium III CPU serial number */
    unsigned long signature;
    eax = 1;
    cpuid(&eax,&ebx,&ecx,&edx);
    signature = eax;

    eax = 3;
    cpuid(&eax,&ebx,&ecx,&edx);
    printf("Processor serial: ");
    printf("%04X",signature >> 16);
    printf("-%04X",signature & 0xffff);
    printf("-%04X",edx >> 16);
    printf("-%04X",edx & 0xffff);
    printf("-%04X",ecx >> 16);
    printf("-%04X\n",ecx & 0xffff);
  }
}

      
/* Decode Intel TLB and cache info descriptors */
void decode_intel_tlb(int x){
  switch(x & 0xff){
  case 0:
    break;
  case 0x1:
    printf("Instruction TLB: 4KB pages, 4-way set assoc, 32 entries\n");
    break;
  case 0x2:
    printf("Instruction TLB: 4MB pages, fully assoc, 2 entries\n");
    break;
  case 0x3:
    printf("Data TLB: 4KB pages, 4-way set assoc, 64 entries\n");
    break;
  case 0x4:
    printf("Data TLB: 4MB pages, 4-way set assoc, 8 entries\n");
    break;
  case 0x6:
    printf("Instruction cache: 8KB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0x8:
    printf("Instruction cache: 16KB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0xa:
    printf("Data cache: 8KB, 2-way set assoc, 32 byte line size\n");
    break;
  case 0xc:
    printf("Data cache: 16KB, 2-way or 4-way set assoc, 32 byte line size\n");
    break;
  case 0x40:
    printf("No L2 cache\n");
    break;
  case 0x41:
    printf("L2 unified cache: 128KB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0x42:
    printf("L2 unified cache: 256KB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0x43:
    printf("L2 unified cache: 512KB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0x44:
    printf("L2 unified cache: 1MB, 4-way set assoc, 32 byte line size\n");
    break;
  case 0x45:
    printf("L2 unified cache: 2MB, 4-way set assoc, 32 byte line size\n");
    break;
  default:
    printf("unknown TLB/cache descriptor: 0x%x\n",x);
    break;
  }
}

/* AMD-specific information */
doamd(){
  int i,maxi,extended;
  unsigned long eax,ebx,ecx,edx;
  int stepping,model,family,reserved,ntlb;

  printf("AMD-specific functions\n");
  eax = 0x80000000;
  cpuid(&eax,&ebx,&ecx,&edx);
  if(eax != 0){
    /* Dump extended info in raw hex */
    maxi = eax;
    
    for(i=0x80000000;i<=maxi;i++){
      eax = i;
      cpuid(&eax,&ebx,&ecx,&edx);
      printf("eax in: 0x%x, eax = %08x ebx = %08x ecx = %08x edx = %08x\n",i,eax,ebx,ecx,edx);
    }
  }

  /* Do standard stuff */
  eax = 0;
  cpuid(&eax,&ebx,&ecx,&edx);
  maxi = eax;

  if(maxi >= 1){
    eax = 1;
    cpuid(&eax,&ebx,&ecx,&edx);
    stepping = eax & 0xf;
    model = (eax >> 4) & 0xf;
    family = (eax >> 8) & 0xf;
    reserved = eax >> 12;

    printf("Family: %d Model: %d [",family,model);
    switch(family){
    case 4:
      printf("486");
      break;
    case 5:
      printf("586/K6");
      break;
    }
    printf("]\n\n");
    printf("Standard feature flags:\n");
    if(edx & (1<<0))
      printf("Floating Point Unit\n");
    if(edx & (1<<1))
      printf("Virtual Mode Extensions\n");
    if(edx & (1<<2))
      printf("Debugging Extensions\n");
    if(edx & (1<<3))
      printf("Page Size Extensions\n");
    if(edx & (1<<4))
      printf("Time Stamp Counter (with RDTSC and CR4 disable bit)\n");
    if(edx & (1<<5))
      printf("Model Specific Registers\n");
    if(edx & (1<<6))
      printf("6 - reserved\n");
    if(edx & (1<<7))
      printf("Machine Check Exception\n");
    if(edx & (1<<8))
      printf("COMPXCHG8B Instruction\n");
    if(edx & (1<<9)){
      if(family == 5 && model == 0)
	printf("Global Paging Extensions\n");
      else
	printf("APIC - On-chip Advanced Programmable Interrupt Controller present and enabled\n");
    }
    if(edx & (1<<10))
      printf("10 - Reserved\n");
    if(edx & (1<<11))
      printf("11 - Reserved\n");
    if(edx & (1<<12))
      printf("Memory Type Range Registers\n");
    if(edx & (1<<13)){
      if(family == 5 && model == 0)
	printf("13 - reserved\n");
      else
	printf("Global Paging Extension\n");
    }
    if(edx & (1<<14))
      printf("14 - reserved\n");
    if(edx & (1<<15))
      printf("Conditional Move Instruction\n");
    if(edx & (1<<16))
      printf("16 - reserved\n");
    if(edx & (1<<17))
      printf("17 - reserved\n");
    if(edx & (1<<18))
      printf("18 - reserved\n");
    if(edx & (1<<19))
      printf("19 - reserved\n");
    if(edx & (1<<20))
      printf("20 - reserved\n");
    if(edx & (1<<21))
      printf("21 - reserved\n");
    if(edx & (1<<22))
      printf("22 - reserved\n");
    if(edx & (1<<23))
      printf("MMX instructions\n");
    if(edx & (1<<24))
      printf("24 - reserved\n");
    if(edx & (1<<25))
      printf("25 - reserved\n");
    if(edx & (1<<26))
      printf("26 - reserved\n");
    if(edx & (1<<27))
      printf("27 - reserved\n");
    if(edx & (1<<28))
      printf("28 - reserved\n");
    if(edx & (1<<29))
      printf("29 - reserved\n");
    if(edx & (1<<30))
      printf("30 - reserved\n");
    if(edx & (1<<31))
      printf("31 - reserved\n");
  }
  /* Check for presence of extended info */
  eax = 0x80000000;
  cpuid(&eax,&ebx,&ecx,&edx);
  if(eax == 0)
    return;

  printf("\nExtended feature flags:\n");
  maxi = eax;
  if(maxi >= 0x80000001){
    eax = 0x80000001;
    cpuid(&eax,&ebx,&ecx,&edx);
    stepping = eax & 0xf;
    model = (eax >> 4) & 0xf;
    family = (eax >> 8) & 0xf;
    reserved = eax >> 12;

    printf("Family: %d Model: %d [",family,model);
    switch(family){
    case 4:
      printf("486");
      break;
    case 5:
      printf("586");
      break;
    case 6:
      printf("K6");
      break;
    }
    printf("]\n");
    printf("Extended feature flags:\n");
    if(edx & (1<<0))
      printf("Floating Point Unit\n");
    if(edx & (1<<1))
      printf("Virtual Mode Extensions\n");
    if(edx & (1<<2))
      printf("Debugging Extensions\n");
    if(edx & (1<<3))
      printf("Page Size Extensions\n");
    if(edx & (1<<4))
      printf("Time Stamp Counter (with RDTSC and CR4 disable bit)\n");
    if(edx & (1<<5))
      printf("K86 family Model Specific Registers with RDMSR & WRMSR\n");
    if(edx & (1<<6))
      printf("6 - reserved\n");
    if(edx & (1<<7))
      printf("Machine Check Exception\n");
    if(edx & (1<<8))
      printf("COMPXCHG8B Instruction\n");
    if(edx & (1<<9))
      printf("9 - reserved\n");
    if(edx & (1<<10))
      printf("10 - Reserved\n");
    if(edx & (1<<11))
      printf("Syscall and Sysret instructions\n");
    if(edx & (1<<12))
      printf("12 - reserved\n");
    if(edx & (1<<13))
      printf("Global paging extension\n");
    if(edx & (1<<14))
      printf("14 - reserved\n");
    if(edx & (1<<15))
      printf("Integer Conditional Move Instruction\n");
    if(edx & (1<<16))
      printf("Floating-point conditional move instructions\n");
    if(edx & (1<<17))
      printf("17 - reserved\n");
    if(edx & (1<<18))
      printf("18 - reserved\n");
    if(edx & (1<<19))
      printf("19 - reserved\n");
    if(edx & (1<<20))
      printf("20 - reserved\n");
    if(edx & (1<<21))
      printf("21 - reserved\n");
    if(edx & (1<<22))
      printf("22 - reserved\n");
    if(edx & (1<<23))
      printf("MMX instructions\n");
    if(edx & (1<<24))
      printf("24 - reserved\n");
    if(edx & (1<<25))
      printf("25 - reserved\n");
    if(edx & (1<<26))
      printf("26 - reserved\n");
    if(edx & (1<<27))
      printf("27 - reserved\n");
    if(edx & (1<<28))
      printf("28 - reserved\n");
    if(edx & (1<<29))
      printf("29 - reserved\n");
    if(edx & (1<<30))
      printf("30 - reserved\n");
    if(edx & (1<<31))
      printf("3DNow instructions\n");
  }
  printf("\n");
  if(maxi >= 0x80000002){
    /* Processor identification string */
    char namestring[49],*cp;
    int j;
    cp = namestring;
    for(j=0x80000002;j<=0x80000004;j++){
      eax = j;
      cpuid(&eax,&ebx,&ecx,&edx);
      
      for(i=0;i<4;i++)
	*cp++ = eax >> (8*i);
      for(i=0;i<4;i++)
	*cp++ = ebx >> (8*i);
      for(i=0;i<4;i++)
	*cp++ = ecx >> (8*i);
      for(i=0;i<4;i++)
	*cp++ = edx >> (8*i);
    }
    *cp++ = '\n';
    printf("Processor name string: %s\n",namestring);
  }
  if(maxi >= 0x80000005){
    /* TLB and cache info */
    eax = 0x80000005;
    cpuid(&eax,&ebx,&ecx,&edx);
    printf("Data TLB: associativity %x #entries %d\n",
	   ebx >> 24,(ebx >> 16) & 0xff);
    printf("Instruction TLB: associativity %x #entries %d\n",
	   (ebx >> 8) & 0xff,ebx & 0xff);
    printf("L1 Data cache: size %d KB associativity %x lines per tag %d line size %d\n",
	   ecx >> 24,(ecx>>16) & 0xff,(ecx >> 8)&0xff,ecx&0xff);

    printf("L1 Instruction cache: size %d KB associativity %x lines per tag %d line size %d\n",
	   edx >> 24,(edx>>16) & 0xff,(edx >> 8)&0xff,edx&0xff);

	   
    
  }
}
